"""WsRpc for Shelly."""

from __future__ import annotations

import asyncio
import contextlib
import hashlib
import logging
import socket
import time
from asyncio import Task, tasks
from collections.abc import Callable, Coroutine, Iterable
from dataclasses import dataclass
from enum import Enum, auto
from http import HTTPStatus
from typing import TYPE_CHECKING, Any

from aiohttp import (
    ClientSession,
    ClientWebSocketResponse,
    WSMessage,
    WSMsgType,
    client_exceptions,
)
from aiohttp.web import (
    Application,
    AppRunner,
    BaseRequest,
    TCPSite,
    WebSocketResponse,
    get,
)
from yarl import URL

from ..const import (
    DEFAULT_HTTP_PORT,
    NOTIFY_WS_CLOSED,
    UNDEFINED,
    WS_API_URL,
    WS_HEARTBEAT,
    UndefinedType,
)
from ..exceptions import (
    ConnectionClosed,
    DeviceConnectionError,
    DeviceConnectionTimeoutError,
    InvalidAuthError,
    InvalidMessage,
    RpcCallError,
)
from ..json import json_bytes, json_loads

_LOGGER = logging.getLogger(__name__)

BUFFER_SIZE = 1024 * 64


class RPCSource(Enum):
    """RPC message source."""

    CLIENT = auto()
    SERVER = auto()


def _receive_json_or_raise(msg: WSMessage) -> dict[str, Any]:
    """Receive json or raise."""
    if msg.type is WSMsgType.TEXT:
        try:
            data: dict[str, Any] = json_loads(msg.data)
        except ValueError as err:
            raise InvalidMessage(f"Received invalid JSON: {msg.data}") from err
        return data

    if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
        raise ConnectionClosed("Connection was closed.")

    if msg.type is WSMsgType.ERROR:
        raise InvalidMessage("Received message error")

    raise InvalidMessage(f"Received non-Text message: {msg.type}")


def hex_hash(message: str) -> str:
    """Get hex representation of sha256 hash of string."""
    return hashlib.sha256(message.encode("utf-8")).hexdigest()


HA2 = hex_hash("dummy_method:dummy_uri")


@dataclass
class AuthData:
    """RPC Auth data class."""

    realm: str
    username: str
    password: str

    def __post_init__(self) -> None:
        """Call after initialization."""
        self.ha1 = hex_hash(f"{self.username}:{self.realm}:{self.password}")

    def get_auth(self, nonce: int | None = None, n_c: int = 1) -> dict[str, Any]:
        """Get auth for RPC calls."""
        cnonce = int(time.time())
        if nonce is None:
            nonce = cnonce - 1800

        # https://shelly-api-docs.shelly.cloud/gen2/Overview/CommonDeviceTraits/#authentication-over-websocket
        hashed = hex_hash(f"{self.ha1}:{nonce}:{n_c}:{cnonce}:auth:{HA2}")

        return {
            "realm": self.realm,
            "username": self.username,
            "nonce": nonce,
            "cnonce": cnonce,
            "response": hashed,
            "algorithm": "SHA-256",
        }


@dataclass
class SessionData:
    """SessionData (src/dst/auth) class."""

    src: str | None
    dst: str | None
    auth: dict[str, Any] | None


class RPCCall:
    """RPCCall class."""

    __slots__ = (
        "auth",
        "call_id",
        "dst",
        "method",
        "params",
        "resolve",
        "result",
        "src",
    )

    def __init__(
        self,
        call_id: int,
        method: str,
        params: dict[str, Any] | None,
        session: SessionData,
        resolve: asyncio.Future[dict[str, Any]],
    ) -> None:
        """Initialize RPC class."""
        self.auth = session.auth
        self.call_id = call_id
        self.params = params
        self.method = method
        self.src = session.src
        self.dst = session.dst
        self.resolve = resolve
        self.result: dict[str, Any] | UndefinedType = UNDEFINED

    def __repr__(self) -> str:
        """Return representation of the call."""
        return (
            "<RPCCall "
            f"method={self.method} "
            f"params={self.params} "
            f"call_id={self.call_id} "
            f"result={self.result}"
            ">"
        )

    @property
    def request_frame(self) -> dict[str, Any]:
        """Request frame."""
        msg = {
            "id": self.call_id,
            "method": self.method,
            "src": self.src,
        }
        for obj in ("params", "dst", "auth"):
            if getattr(self, obj) is not None:
                msg[obj] = getattr(self, obj)
        return msg


class WsBase:
    """Base class for WebSocket handlers."""

    def __init__(self) -> None:
        """Initialize WsBase class."""
        self._background_tasks: set[Task] = set()

    def _create_and_track_task(self, func: Coroutine) -> None:
        """Create and and hold strong reference to the task."""
        task = asyncio.create_task(func)
        self._background_tasks.add(task)
        task.add_done_callback(self._background_tasks.discard)


class WsRPC(WsBase):
    """WsRPC class."""

    def __init__(
        self,
        ip_address: str,
        on_notification: Callable[[RPCSource, str, dict | None], None],
        port: int = DEFAULT_HTTP_PORT,
    ) -> None:
        """Initialize WsRPC class."""
        super().__init__()
        self._auth_data: AuthData | None = None
        self._ip_address = ip_address
        self._port = port
        self._on_notification = on_notification
        self._rx_task: tasks.Task[None] | None = None
        self._client: ClientWebSocketResponse | None = None
        self._calls: dict[int, RPCCall] = {}
        self._call_id = 0
        self._session = SessionData(f"aios-{id(self)}", None, None)
        self._loop = asyncio.get_running_loop()

    @property
    def _next_id(self) -> int:
        self._call_id += 1
        return self._call_id

    async def connect(self, aiohttp_session: ClientSession) -> None:
        """Connect to device."""
        if self.connected:
            raise RuntimeError("Already connected")

        _LOGGER.debug("Trying to connect to device at %s", self._ip_address)
        try:
            self._client = await aiohttp_session.ws_connect(
                URL.build(
                    scheme="http", host=self._ip_address, port=self._port, path="/rpc"
                ),
                heartbeat=WS_HEARTBEAT,
            )
        except (
            client_exceptions.WSServerHandshakeError,
            client_exceptions.ClientError,
        ) as err:
            raise DeviceConnectionError(err) from err

        # Try to reduce the pressure on shelly device as it measures
        # ram in bytes and we measure ram in megabytes.
        sock: socket.socket = self._client.get_extra_info("socket")
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, BUFFER_SIZE)
        except OSError as err:
            _LOGGER.warning(
                "%s:%s: Failed to set socket receive buffer size: %s",
                self._ip_address,
                self._port,
                err,
            )

        self._rx_task = asyncio.create_task(self._rx_msgs())

        _LOGGER.info("Connected to %s", self._ip_address)

    async def disconnect(self) -> None:
        """Disconnect all sessions."""
        if self._rx_task is not None:
            self._rx_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._rx_task
            self._rx_task = None
        if self._client is None:
            return

        await self._client.close()

    def set_auth_data(self, realm: str, username: str, password: str) -> None:
        """Set authentication data and generate session auth."""
        self._auth_data = AuthData(realm, username, password)
        self._session.auth = self._auth_data.get_auth()

    async def _handle_call(self, frame_id: str) -> None:
        if TYPE_CHECKING:
            assert self._client

        await self._send_json(
            {
                "id": frame_id,
                "src": self._session.src,
                "error": {"code": 500, "message": "Not Implemented"},
            }
        )

    def handle_frame(self, source: RPCSource, frame: dict[str, Any]) -> None:
        """Handle RPC frame."""
        if peer_src := frame.get("src"):
            if self._session.dst is not None and peer_src != self._session.dst:
                _LOGGER.warning(
                    "Remote src changed: %s -> %s", self._session.dst, peer_src
                )
            self._session.dst = peer_src

        frame_id = frame.get("id")

        if method := frame.get("method"):
            # peer is invoking a method
            params = frame.get("params")
            if frame_id:
                # and expects a response
                _LOGGER.debug("handle call for frame_id: %s", frame_id)
                self._create_and_track_task(self._handle_call(frame_id))
            else:
                # this is a notification
                _LOGGER.debug("Notification: %s %s", method, params)
                try:
                    self._on_notification(source, method, params)
                except Exception as err:
                    _LOGGER.exception(
                        "Error handling notification frame: %s", frame, exc_info=err
                    )

            return

        if frame_id:
            # looks like a response
            if (call := self._calls.pop(frame_id, None)) is None:
                _LOGGER.warning(
                    "Response from (%s:%s) for an unknown request id: %s: %s",
                    self._ip_address,
                    self._port,
                    frame_id,
                    method,
                )
                return

            if not call.resolve.cancelled():
                call.resolve.set_result(frame)

            return

        _LOGGER.warning("Invalid frame: %s", frame)

    async def _rx_msgs(self) -> None:
        if TYPE_CHECKING:
            assert self._client

        try:
            while True:
                try:
                    msg = await self._client.receive()
                    frame = _receive_json_or_raise(msg)
                    _LOGGER.debug(
                        "recv(%s:%s): %s", self._ip_address, self._port, frame
                    )
                except InvalidMessage as err:
                    _LOGGER.error(
                        "Invalid Message from host %s:%s: %s",
                        self._ip_address,
                        self._port,
                        err,
                    )
                except (ConnectionClosed, client_exceptions.ClientConnectionResetError):
                    _LOGGER.debug(
                        "Connection issue with device %s:%s",
                        self._ip_address,
                        self._port,
                    )
                    break
                except Exception:
                    _LOGGER.exception("Unexpected error while receiving message")
                    raise

                if self._client.closed:
                    break

                self.handle_frame(RPCSource.CLIENT, frame)
        finally:
            _LOGGER.debug(
                "Websocket client connection from %s:%s closed",
                self._ip_address,
                self._port,
            )

            for call_item in self._calls.values():
                if not call_item.resolve.done():
                    call_item.resolve.set_exception(DeviceConnectionError(call_item))
            self._calls.clear()

            client = self._client
            self._client = None
            self._on_notification(RPCSource.CLIENT, NOTIFY_WS_CLOSED, None)

            # Close last since the await can yield
            # to the event loop and we want to minimize
            # race conditions
            if not client.closed:
                await client.close()

    @property
    def connected(self) -> bool:
        """Return if we're currently connected."""
        return self._client is not None and not self._client.closed

    async def call(
        self,
        method: str,
        params: dict[str, Any] | None = None,
        timeout: int = 10,
    ) -> dict[str, Any]:
        """Websocket RPC call."""
        return (await self.calls([(method, params)], timeout))[0]

    def _raise_for_unrecoverable_errors(
        self, resp: dict[str, Any], allow_auth_retry: bool
    ) -> None:
        """Raise for unrecoverable errors."""
        try:
            error = resp["error"]
            code = error["code"]
            msg = error["message"]
        except KeyError as err:
            raise RpcCallError(0, f"bad response: {resp}") from err

        if code != HTTPStatus.UNAUTHORIZED.value:
            raise RpcCallError(code, msg)

        if allow_auth_retry and self._auth_data is not None:
            return

        raise InvalidAuthError(msg)

    async def calls(
        self, calls: Iterable[tuple[str, dict[str, Any] | None]], timeout: float = 10.0
    ) -> list[dict[str, Any]]:
        """Websocket RPC calls."""
        # Try request with initial/last call auth data
        all_successful, results = await self._rpc_calls(calls, timeout)
        if all_successful:
            # If all_successful, return results immediately
            # mypy does not know that .result is never
            # None when all_successful is True so we need
            # to ignore the type check here
            return [call.result for call in results]  # type: ignore[misc]

        # Partial success, try to update auth and retry
        to_retry: list[RPCCall] = []
        successful: list[dict[str, Any]] = []
        for call in results:
            if (result := call.result) is not UNDEFINED:
                successful.append(result)
                continue
            resp = call.resolve.result()
            self._raise_for_unrecoverable_errors(resp, allow_auth_retry=True)
            if not to_retry:
                # Update auth from response and try with new auth data
                # If we have multiple calls, we only need to update auth once
                if TYPE_CHECKING:
                    # _raise_for_unrecoverable_errors ensures that auth_data is not None
                    assert self._auth_data is not None
                auth = json_loads(resp["error"]["message"])
                self._session.auth = self._auth_data.get_auth(
                    auth["nonce"], auth.get("nc", 1)
                )
            to_retry.append(call)

        _, results = await self._rpc_calls(
            [(call.method, call.params) for call in to_retry], timeout
        )
        for call in results:
            if (result := call.result) is UNDEFINED:
                resp = call.resolve.result()
                self._raise_for_unrecoverable_errors(resp, allow_auth_retry=False)
            else:
                successful.append(result)

        return successful

    async def _rpc_calls(
        self, rpc_calls: Iterable[tuple[str, dict[str, Any] | None]], timeout: float
    ) -> tuple[bool, list[RPCCall]]:
        """Websocket RPC call.

        calls is a tuple of tuples of
        (
           (method, params),
           ...

        """
        if self._client is None:
            raise RuntimeError("Not connected")

        sent_calls: list[RPCCall] = []
        loop = self._loop
        all_successful: bool = True
        future: asyncio.Future[dict[str, Any]]

        try:
            async with asyncio.timeout(timeout):
                for method, params in rpc_calls:
                    call_id = self._next_id
                    future = loop.create_future()
                    call = RPCCall(call_id, method, params, self._session, future)
                    sent_calls.append(call)
                    self._calls[call_id] = call
                    await self._send_json(call.request_frame)

                # Wait for all the responses
                for call in sent_calls:
                    response = await call.resolve
                    if "result" not in response:
                        all_successful = False
                        continue
                    call.result = response["result"]
        except TimeoutError as exc:
            for call in sent_calls:
                with contextlib.suppress(asyncio.CancelledError):
                    call.resolve.cancel()
                    await call.resolve
                # Ensure the call is removed from the calls dict
                # on failure
                self._calls.pop(call.call_id, None)
            raise DeviceConnectionTimeoutError(sent_calls) from exc

        if _LOGGER.isEnabledFor(logging.DEBUG):
            for call in sent_calls:
                _LOGGER.debug(
                    "result(%s:%s): %s(%s) -> %s",
                    self._ip_address,
                    self._port,
                    call.method,
                    call.params,
                    call.result,
                )

        return all_successful, sent_calls

    async def _send_json(self, data: dict[str, Any]) -> None:
        """Send json frame to device."""
        _LOGGER.debug("send(%s:%s): %s", self._ip_address, self._port, data)

        if TYPE_CHECKING:
            assert self._client

        await self._client.send_frame(json_bytes(data), WSMsgType.TEXT)


class WsServer(WsBase):
    """WsServer class."""

    def __init__(self) -> None:
        """Initialize WsServer class."""
        super().__init__()
        self._runner: AppRunner | None = None
        self.subscriptions: dict[str, Callable] = {}

    async def initialize(self, port: int, api_url: str = WS_API_URL) -> None:
        """Initialize the websocket server, used only in standalone mode."""
        app = Application()
        app.add_routes([get(api_url, self.websocket_handler)])
        self._runner = AppRunner(app)
        await self._runner.setup()
        site = TCPSite(self._runner, port=port)
        await site.start()

    def close(self) -> None:
        """Stop the websocket server."""
        if self._runner is not None:
            self._create_and_track_task(self._runner.cleanup())

    async def websocket_handler(self, request: BaseRequest) -> WebSocketResponse:
        """Handle connections from sleeping devices."""
        ip = request.remote
        _LOGGER.debug("Websocket server connection from %s starting", ip)
        ws_res = WebSocketResponse(protocols=["json-rpc"])
        await ws_res.prepare(request)
        _LOGGER.debug("Websocket server connection from %s ready", ip)

        async for msg in ws_res:
            try:
                frame = _receive_json_or_raise(msg)
                _LOGGER.debug("recv(%s): %s", ip, frame)
            except ConnectionClosed:
                await ws_res.close()
            except InvalidMessage as err:
                _LOGGER.debug("Invalid Message from host %s: %s", ip, err)
            else:
                try:
                    device_id = frame["src"].split("-")[1].upper()
                except (KeyError, IndexError) as err:
                    _LOGGER.debug("Invalid device id from host %s: %s", ip, err)
                    continue

                if device_id in self.subscriptions:
                    _LOGGER.debug(
                        "Calling WsRPC message update for device id %s", device_id
                    )
                    self.subscriptions[device_id](frame)
                    continue

                if ip in self.subscriptions:
                    _LOGGER.debug("Calling WsRPC message update for host %s", ip)
                    self.subscriptions[ip](frame)

        _LOGGER.debug("Websocket server connection from %s closed", ip)
        return ws_res

    def subscribe_updates(self, ip: str, message_received: Callable) -> Callable:
        """Subscribe to received updates."""
        _LOGGER.debug("Adding device %s to WsServer message subscriptions", ip)
        self.subscriptions[ip] = message_received
        return lambda: self.subscriptions.pop(ip)



"""Tests for JSON helper."""

import pytest

from aioshelly.json import json_bytes, json_dumps


def test_json_bytes_with_dict() -> None:
    """Test json_bytes with a dict and non-str keys."""
    data = {"id": 1, "method": "Shelly.GetDeviceInfo", 3: "test"}
    result = json_bytes(data)

    assert isinstance(result, bytes)
    assert result == b'{"id":1,"method":"Shelly.GetDeviceInfo","3":"test"}'


def test_json_bytes_with_set() -> None:
    """Test json_bytes with a set."""
    data = {1, 2}
    result = json_bytes(data)

    assert isinstance(result, bytes)
    assert result in [b"[1,2]", b"[2,1]"]


def test_json_bytes_with_tuple() -> None:
    """Test json_bytes with a tuple."""
    data = (1, 2, 3)
    result = json_bytes(data)

    assert isinstance(result, bytes)
    assert result == b"[1,2,3]"


def test_json_bytes_with_non_serializable() -> None:
    """Test json_bytes with non-serializable object."""
    with pytest.raises(TypeError):
        json_bytes(object())


def test_json_dumps_with_dict() -> None:
    """Test json_dumps with a dict and non-str keys."""
    data = {"id": 1, "method": "Shelly.GetDeviceInfo", 3: "test"}
    result = json_dumps(data)

    assert isinstance(result, str)
    assert result == '{"id":1,"method":"Shelly.GetDeviceInfo","3":"test"}'


def test_json_dumps_with_set() -> None:
    """Test json_dumps with a set."""
    data = {1, 2}
    result = json_dumps(data)

    assert isinstance(result, str)
    assert result in ["[1,2]", "[2,1]"]


def test_json_dumps_with_tuple() -> None:
    """Test json_dumps with a tuple."""
    data = (1, 2, 3)
    result = json_dumps(data)

    assert isinstance(result, str)
    assert result == "[1,2,3]"


def test_json_dumps_with_non_serializable() -> None:
    """Test json_dumps with non-serializable object."""
    with pytest.raises(TypeError):
        json_dumps(object())



"""Tests for RPC device."""

import asyncio
import pathlib
from typing import Any

from orjson import loads


def get_device_fixture_path(device: str, filename: str) -> pathlib.Path:
    """Get path of a device fixture."""
    return pathlib.Path(__file__).parent.joinpath("fixtures", device, filename)


async def load_device_fixture(device: str, filename: str) -> dict[str, Any]:
    """Load a device fixture."""
    fixture_path = get_device_fixture_path(device, filename)
    json_bytes = await asyncio.get_running_loop().run_in_executor(
        None, fixture_path.read_bytes
    )
    return loads(json_bytes)



from aioshelly import rpc_device


def test_exports() -> None:
    """Test objects are available at top level of rpc_device."""
    assert hasattr(rpc_device, "bluetooth_mac_from_primary_mac")
    assert hasattr(rpc_device, "RpcDevice")
    assert hasattr(rpc_device, "RpcUpdateType")
    assert hasattr(rpc_device, "WsServer")



"""WiFi provisioning via BLE for Shelly devices."""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, cast

from ..common import ConnectionOptions
from ..rpc_device import RpcDevice

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from bleak.backends.device import BLEDevice


@asynccontextmanager
async def ble_rpc_device(ble_device: BLEDevice) -> AsyncIterator[RpcDevice]:
    """Create and manage BLE RPC device connection.

    Args:
        ble_device: BLE device to connect to

    Yields:
        Initialized RPC device

    Raises:
        DeviceConnectionError: If connection to device fails

    """
    options = ConnectionOptions(ble_device=ble_device)
    device = await RpcDevice.create(
        aiohttp_session=None,
        ws_context=None,
        ip_or_options=options,
    )

    try:
        await device.initialize()
        yield device
    finally:
        await device.shutdown()


async def async_scan_wifi_networks(ble_device: BLEDevice) -> list[dict[str, Any]]:
    """Scan for WiFi networks via BLE.

    Args:
        ble_device: BLE device to connect to

    Returns:
        List of WiFi networks with ssid, rssi, auth fields

    Raises:
        DeviceConnectionError: If connection to device fails
        RpcCallError: If RPC call fails

    """
    async with ble_rpc_device(ble_device) as device:
        # WiFi scan can take up to 20 seconds - use 30s timeout to be safe
        scan_result = await device.call_rpc("WiFi.Scan", timeout=30)
        return cast(list[dict[str, Any]], scan_result.get("results", []))


async def async_provision_wifi(ble_device: BLEDevice, ssid: str, password: str) -> None:
    """Provision WiFi credentials to device via BLE.

    Args:
        ble_device: BLE device to connect to
        ssid: WiFi network SSID
        password: WiFi network password

    Raises:
        DeviceConnectionError: If connection to device fails
        RpcCallError: If RPC call fails

    """
    async with ble_rpc_device(ble_device) as device:
        await device.call_rpc(
            "WiFi.SetConfig",
            {
                "config": {
                    "sta": {
                        "ssid": ssid,
                        "pass": password,
                        "enable": True,
                    }
                }
            },
        )



# Run with python3 example.py -h for help
"""aioshelly usage example."""

from __future__ import annotations

import argparse
import asyncio
import json
import logging
import traceback
from functools import partial
from pathlib import Path

from aiohttp import ClientSession
from common import (
    check_rpc_device_supports_scripts,
    close_connections,
    coap_context,
    connect_and_print_device,
    create_device,
    device_updated,
    init_device,
    print_device,
    update_outbound_ws,
    wait_for_keyboard_interrupt,
    ws_context,
)

from aioshelly.common import ConnectionOptions
from aioshelly.const import (
    BLOCK_GENERATIONS,
    DEFAULT_HTTP_PORT,
    RPC_GENERATIONS,
    WS_API_URL,
)


async def test_single(options: ConnectionOptions, init: bool, gen: int | None) -> None:
    """Test single device."""
    async with ClientSession() as aiohttp_session:
        device = await create_device(aiohttp_session, options, gen)

        if init and not await init_device(device):
            return

        print_device(device)

        device.subscribe_updates(partial(device_updated, action=print_device))

        await wait_for_keyboard_interrupt()

        await device.shutdown()
        close_connections()


async def test_devices(init: bool, gen: int | None) -> None:
    """Test multiple devices."""
    options: ConnectionOptions

    with Path.open(Path("devices.json"), encoding="utf8") as fp:
        device_options = [ConnectionOptions(**json.loads(line)) for line in fp]

    async with ClientSession() as aiohttp_session:
        results = await asyncio.gather(
            *[
                asyncio.gather(
                    connect_and_print_device(aiohttp_session, options, init, gen),
                )
                for options in device_options
            ],
            return_exceptions=True,
        )

        for options, result in zip(device_options, results, strict=False):
            if isinstance(result, bool) and not result:
                print(f"Error printing device @ {options.ip_address}:{options.port}")
            elif isinstance(result, Exception):
                print()
                traceback.print_tb(result.__traceback__)
                print(result)

        await wait_for_keyboard_interrupt()
        close_connections()


def get_arguments() -> tuple[argparse.ArgumentParser, argparse.Namespace]:
    """Get parsed passed in arguments."""
    parser = argparse.ArgumentParser(description="aioshelly example")
    parser.add_argument(
        "--ip_address", "-ip", type=str, help="Test single device by IP address"
    )
    parser.add_argument(
        "--device_port",
        "-dp",
        type=int,
        default=DEFAULT_HTTP_PORT,
        help="Port to use when testing single device",
    )
    parser.add_argument(
        "--coap_port",
        "-cp",
        type=int,
        default=5683,
        help="Specify CoAP UDP port (default=5683)",
    )
    parser.add_argument(
        "--ws_port",
        "-wp",
        type=int,
        default=8123,
        help="Specify WebSocket TCP port (default=8123)",
    )
    parser.add_argument(
        "--ws_api_url",
        "-au",
        type=str,
        default=WS_API_URL,
        help=f"Specify WebSocket API URL (default={WS_API_URL})",
    )
    parser.add_argument(
        "--devices",
        "-d",
        action="store_true",
        help=(
            'Connect to all the devices in "devices.json" '
            "at once and print their status"
        ),
    )
    parser.add_argument(
        "--init", "-i", action="store_true", help="Init device(s) at startup"
    )
    parser.add_argument("--username", "-u", type=str, help="Set device username")
    parser.add_argument("--password", "-p", type=str, help="Set device password")

    gen = parser.add_mutually_exclusive_group()
    gen.add_argument(
        "--gen1",
        "-g1",
        action="store_const",
        const=1,
        dest="gen",
        help="Force Gen1 (CoAP) device",
    )
    gen.add_argument(
        "--gen2",
        "-g2",
        action="store_const",
        const=2,
        dest="gen",
        help="Force Gen 2 (RPC) device",
    )
    gen.add_argument(
        "--gen3",
        "-g3",
        action="store_const",
        const=3,
        dest="gen",
        help="Force Gen 3 (RPC) device",
    )
    gen.add_argument(
        "--gen4",
        "-g4",
        action="store_const",
        const=4,
        dest="gen",
        help="Force Gen 4 (RPC) device",
    )

    parser.add_argument(
        "--debug", "-deb", action="store_true", help="Enable debug level for logging"
    )
    parser.add_argument(
        "--mac", "-m", type=str, help="Optional device MAC to subscribe for updates"
    )
    parser.add_argument(
        "--update_ws",
        "-uw",
        type=str,
        help="Update outbound WebSocket (for RPC device) and exit",
    )
    parser.add_argument(
        "--listen_ip_address",
        "-lip",
        type=str,
        nargs="*",
        default=None,
        help="Listen ip address for incoming CoAP packets",
    )
    parser.add_argument(
        "--supports_scripts",
        "-ss",
        action="store_true",
        help="Check if device supports scripts",
    )

    arguments = parser.parse_args()

    return parser, arguments


async def main() -> None:
    """Run main."""
    parser, args = get_arguments()

    await coap_context.initialize(args.coap_port, args.listen_ip_address)
    await ws_context.initialize(args.ws_port, args.ws_api_url)

    if not args.init and not args.gen:
        parser.error("specify gen if no device init at startup")

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
        # if gen is in args reduce logging for other gens
        if args.gen in BLOCK_GENERATIONS:
            logging.getLogger("aioshelly.rpc_device").setLevel(logging.INFO)
        elif args.gen in RPC_GENERATIONS:
            logging.getLogger("aioshelly.block_device").setLevel(logging.INFO)

    if args.devices:
        await test_devices(args.init, args.gen)
    elif args.ip_address:
        if args.username and args.password is None:
            parser.error("--username and --password must be used together")
        options = ConnectionOptions(
            args.ip_address,
            args.username,
            args.password,
            device_mac=args.mac,
            port=args.device_port,
        )
        if args.update_ws:
            await update_outbound_ws(options, args.init, args.update_ws)
        elif args.supports_scripts:
            await check_rpc_device_supports_scripts(options, args.gen)
        else:
            await test_single(options, args.init, args.gen)
    else:
        parser.error("--ip_address or --devices must be specified")


if __name__ == "__main__":
    asyncio.run(main())



"""Shelly Gen2 RPC based device models."""

from __future__ import annotations

from typing import TypedDict


class ShellyScript(TypedDict, total=False):
    """Shelly Script."""

    id: int
    name: str
    enable: bool
    running: bool


class ShellyScriptCode(TypedDict, total=False):
    """Shelly Script Code."""

    data: str


class ShellyBLERpcConfig(TypedDict, total=False):
    """Shelly BLE RPC Config."""

    enable: bool


class ShellyBLEConfig(TypedDict, total=False):
    """Shelly BLE Config."""

    enable: bool
    rpc: ShellyBLERpcConfig


class ShellyBLESetConfig(TypedDict, total=False):
    """Shelly BLE Set Config."""

    restart_required: bool


class ShellyWsConfig(TypedDict, total=False):
    """Shelly Outbound Websocket Config."""

    enable: bool
    server: str | None
    ssl_ca: str


class ShellyWsSetConfig(TypedDict, total=False):
    """Shelly Outbound Websocket Set Config."""

    restart_required: bool



"""Common code for Shelly library."""

from __future__ import annotations

import asyncio
import ipaddress
import logging
from dataclasses import dataclass
from socket import gethostbyname
from typing import TYPE_CHECKING, Any

from aiohttp import BasicAuth, ClientSession, ClientTimeout
from yarl import URL

if TYPE_CHECKING:
    from bleak import BLEDevice

from .const import (
    CONNECT_ERRORS,
    DEFAULT_HTTP_PORT,
    DEVICE_IO_TIMEOUT,
    DEVICES,
    FIRMWARE_PATTERN,
    MIN_FIRMWARE_DATES,
)
from .exceptions import (
    DeviceConnectionError,
    DeviceConnectionTimeoutError,
    InvalidHostError,
    MacAddressMismatchError,
    ShellyError,
)

_LOGGER = logging.getLogger(__name__)

DEVICE_IO_TIMEOUT_CLIENT_TIMEOUT = ClientTimeout(total=DEVICE_IO_TIMEOUT)


@dataclass
class ConnectionOptions:
    """Shelly options for connection."""

    ip_address: str | None = None
    username: str | None = None
    password: str | None = None
    temperature_unit: str = "C"
    auth: BasicAuth | None = None
    device_mac: str | None = None
    port: int = DEFAULT_HTTP_PORT
    ble_device: BLEDevice | None = None

    def __post_init__(self) -> None:
        """Call after initialization."""
        if self.ip_address is None and self.ble_device is None:
            raise ValueError("Must provide either ip_address or ble_device")

        if self.ip_address is not None and self.ble_device is not None:
            raise ValueError("Cannot provide both ip_address and ble_device")

        if self.username is not None:
            if self.password is None:
                raise ValueError("Supply both username and password")

            object.__setattr__(self, "auth", BasicAuth(self.username, self.password))


IpOrOptionsType = str | ConnectionOptions


async def process_ip_or_options(ip_or_options: IpOrOptionsType) -> ConnectionOptions:
    """Return ConnectionOptions class from ip str or ConnectionOptions."""
    if isinstance(ip_or_options, str):
        options = ConnectionOptions(ip_address=ip_or_options)
    else:
        options = ip_or_options

    # Only process IP address if provided (not for BLE connections)
    if options.ip_address is not None:
        try:
            ipaddress.ip_address(options.ip_address)
        except ValueError:
            loop = asyncio.get_running_loop()
            options.ip_address = await loop.run_in_executor(
                None, gethostbyname, options.ip_address
            )

    return options


async def get_info(
    aiohttp_session: ClientSession,
    ip_address: str,
    device_mac: str | None = None,
    port: int = DEFAULT_HTTP_PORT,
) -> dict[str, Any]:
    """Get info from device through REST call."""
    error: ShellyError
    try:
        async with aiohttp_session.get(
            URL.build(scheme="http", host=ip_address, port=port, path="/shelly"),
            raise_for_status=True,
            timeout=DEVICE_IO_TIMEOUT_CLIENT_TIMEOUT,
        ) as resp:
            result: dict[str, Any] = await resp.json()
    except TimeoutError as err:
        error = DeviceConnectionTimeoutError(err)
        _LOGGER.debug("host %s:%s: timeout error: %r", ip_address, port, error)
        raise error from err
    except ValueError as err:
        error = InvalidHostError(err)
        _LOGGER.debug("host %s is invalid: %r", ip_address, error)
        raise error from err
    except CONNECT_ERRORS as err:
        error = DeviceConnectionError(err)
        _LOGGER.debug("host %s:%s: error: %r", ip_address, port, error)
        raise error from err

    mac = result["mac"]
    if device_mac and device_mac != mac:
        error = MacAddressMismatchError(f"Input MAC: {device_mac}, Shelly MAC: {mac}")
        _LOGGER.debug("host %s:%s: error: %r", ip_address, port, error)
        raise error

    return result


def is_firmware_supported(gen: int, model: str, firmware_version: str) -> bool:
    """Return True if firmware is supported."""
    fw_ver: int | None
    if device := DEVICES.get(model):
        # Specific model is known
        if not device.supported:
            return False
        fw_ver = device.min_fw_date
    elif not (fw_ver := MIN_FIRMWARE_DATES.get(gen)):
        # Protection against future generations of devices.
        return False

    match = FIRMWARE_PATTERN.search(firmware_version)
    if match is None:
        return False
    # We compare firmware release dates because Shelly version numbering is
    # inconsistent, sometimes the word is used as the version number.
    return int(match[0]) >= fw_ver



from aioshelly.rpc_device.utils import bluetooth_mac_from_primary_mac


def test_bluetooth_mac_from_primary_mac() -> None:
    """Test bluetooth_mac_from_primary_mac."""
    assert bluetooth_mac_from_primary_mac("0A1B2C3D4E5F") == "0A1B2C3D4E61"
    assert bluetooth_mac_from_primary_mac("0A1B2C3D4EA0") == "0A1B2C3D4EA2"
    assert bluetooth_mac_from_primary_mac("0A1B2C3D4EF0") == "0A1B2C3D4EF2"
    assert bluetooth_mac_from_primary_mac("0A1B2C3D4EC9") == "0A1B2C3D4ECB"
